""" conversation state definition """
import logging
import typing as ta
import uuid

from asyncio import Queue, Event

from pydantic import BaseModel


logger = logging.getLogger("agt")


class Message(BaseModel):
    text: str
    image_url: ta.Optional[str]
    ssml: ta.Optional[str]


def generate_session_id():
    return str(uuid.uuid4())


class Entry:
    def __init__(
        self,
        text: str = None,
        image_url: ta.Optional[str] = None,
        ssml: ta.Optional[str] = None,
    ):
        self.text = text
        self.image_url = image_url
        self.ssml = ssml

    def __repr__(self):
        if self.image_url:
            return " #img#: ".join([self.text or "", self.image_url])
        return self.text or ""


class BotEntry(Entry):
    def __repr__(self):
        return f"BOT >> {super().__repr__()}"


class UserEntry(Entry):
    def __repr__(self):
        return f"USER >> {super().__repr__()}"


class ConversationState:
    def __init__(self, output_callback):
        self.output_callback = output_callback
        self.session_id: str = generate_session_id()
        self.log: ta.List[Entry] = []
        self.out_of_context_handlers = []
        self._inputs_queue = None
        self._bot_wait_for_input_event = Event()

        self.memory = {}

    async def say(
        self,
        text: str,
        image_url: ta.Optional[str] = None,
        ssml: ta.Optional[str] = None,
    ):
        """Utter text message

        Arguments:
            text {str} -- The text message
        """
        logger.debug(f"BOT:{self.session_id}: {text}")

        if image_url:
            logger.debug(f"BOT:{self.session_id}: {image_url}")

        self.log.append(BotEntry(text=text, image_url=image_url, ssml=ssml))

        await self.output_callback(Message(text=text, image_url=image_url, ssml=ssml))

    async def put_user_input(self, user_input: str):
        if not self._inputs_queue:
            self._inputs_queue = Queue()

        await self._inputs_queue.put(user_input)

    async def user_input(self) -> str:
        """Get next user input / wait for it

        Returns:
            str -- The user input
        """
        if not self._inputs_queue:
            # lazy load queue
            self._inputs_queue = Queue()
        if self._inputs_queue.empty():
            self._bot_wait_for_input_event.set()
        user_input = await self._inputs_queue.get()
        logger.debug(f"USER:{self.session_id}: {user_input}")
        self.log.append(UserEntry(user_input))
        return user_input

    def last_user_input(self) -> ta.Optional[str]:
        last_user_input = next(
            filter(lambda e: isinstance(e, UserEntry), reversed(self.log)), None
        )
        if last_user_input:
            return last_user_input.text
        return None

    def set_out_of_context_handler(self, handler: ta.Callable):
        self.out_of_context_handlers.append(handler)

    def pop_out_of_context_handler(self):
        self.out_of_context_handlers.pop()

    async def out_of_context(self, user_input: str, *args, **kwargs):
        if len(self.out_of_context_handlers) > 0:
            await self.out_of_context_handlers[-1](self, user_input, *args, **kwargs)

    async def bot_listen(self):
        await self._bot_wait_for_input_event.wait()
        self._bot_wait_for_input_event.clear()


class OutOfContext:
    def __init__(self, state, handler):
        self.state = state
        self.state.set_out_of_context_handler(handler)

    def __enter__(self):
        return self.state

    def __exit__(self, type, value, traceback):
        self.state.pop_out_of_context_handler()


class Outputs:
    def __init__(self, success=True, **kwargs):
        self.success = success
        self.outputs = kwargs



import typing
from agt.nlu.word_regex import Intent, Pattern, WILDCARD, AnyWords


yes_intent = Intent(
    Pattern(WILDCARD, ("yes", "yup", "sure", "yep", "yeah", "yea"), WILDCARD),
    Pattern("y"),
)

no_intent = Intent(
    Pattern(WILDCARD, ("no", "nope", "nop", "never"), WILDCARD), Pattern("n")
)


async def say_component(state, text):
    await state.say(text)


def pack_followup_action(state, followup_action):
    if isinstance(followup_action, typing.Awaitable):
        return followup_action
    elif isinstance(followup_action, str):
        return say_component(state, followup_action)

    raise ValueError("Invalid follow-up action")


def followup(intent, followup_action):
    return intent, followup_action


def keywords(keywords_list, followup_action):
    return Intent(Pattern(WILDCARD, keywords_list, WILDCARD)), followup_action


def followups(*followups_intent_actions, yes=None, no=None, default=None):
    async def followup_comp(state, user_input=None):
        if not user_input:
            user_input = await state.user_input()

        followup_actions = []
        if yes:
            followup_actions.append((yes_intent, yes))
        if no:
            followup_actions.append((no_intent, no))

        followup_actions.extend(followups_intent_actions)

        for intent, followup_action in followup_actions:
            if intent(user_input):
                return await pack_followup_action(state, followup_action)

        if default:
            return await pack_followup_action(state, default)

    return followup_comp


async def pick_first_match(user_input, intent_map, default=None):
    for intent, (callback, *args) in intent_map.items():
        if intent(user_input):
            return await callback(*args)
    if default:
        func, *args = default
        return await func(*args)



"""
    Compile simple word patterns to regex

    Examples:
    intent = Intent(
        Pattern("the", "boy", "ate", "an", "apple")
    )
    intent("the boy ate an apple") -> True
    intent("the boy ate an orange") -> False

    intent = Intent(
        Pattern("the", "boy", "ate", "an", AnyWords(min=1, max=1))
    )
    intent("the boy ate an apple") -> True
    intent("the boy ate an orange") -> True

    intent = Intent(
        Pattern("the", Words("boy", "girl"), "ate", "an", AnyWords(min=1, max=1))
    )
    intent("the boy ate an apple") -> True
    intent("the boy ate an orange") -> True
    intent("the girl ate an orange") -> True
    intent("the girl ate a banana") -> False

    intent = Intent(
        Pattern("the", ("boy", "girl"), "ate", WordsRegex(r"an?"), AnyWords(min=1, max=1))
    )
    intent("the boy ate an apple") -> True
    intent("the boy ate an orange") -> True
    intent("the girl ate an orange") -> True
    intent("the girl ate a banana") -> True
    intent("a nice boy ate an apple") -> False

    intent = Intent(
        Pattern(WILDCARD, Words("boy", "girl"), "ate", WordsRegex(r"an?"), AnyWords(min=1, max=1))
    )
    intent("a nice boy ate an apple") -> True

"""
import typing as ta
import abc
import re

from collections import defaultdict
from typing import Optional


class PatternElement(abc.ABC):
    """
    Base class for element in pattern
    """

    def __init__(self, name=None) -> None:
        self.name = name

    @abc.abstractmethod
    def regex_transformation(self):
        raise NotImplementedError


class RegexElement(PatternElement):
    """
    element for any regex in a word position
    """

    def __init__(self, pattern, **kwargs):
        super().__init__(**kwargs)
        self.pattern = pattern

    def regex_transformation(self):
        return self.pattern


class WordsRegex(PatternElement):
    """
    List of word regex patterns to match against a single word position
    """

    def __init__(self, *patterns, **kwargs):
        super().__init__(**kwargs)
        self.word_list_tran = f"\\b({'|'.join(patterns)})\\b"

    def regex_transformation(self):
        return self.word_list_tran


class Words(WordsRegex):
    """
    List of words to match against a single word position
    """

    def __init__(self, *words, **kwargs):
        super().__init__(**kwargs)
        self.word_list_tran = f"\\b({'|'.join(re.escape(w) for w in words)})\\b"


class AnyWords(PatternElement):
    """
    Any word configurable with min, max
    """

    def __init__(self, min="", max="", **kwargs):
        super().__init__(**kwargs)
        self.pattern = "(\\s?\\b\\w+\\b\\s?){" + f"{str(min)},{str(max)}" + "}"

    def regex_transformation(self):
        return self.pattern


class Wildcard(RegexElement):
    def __init__(self, **kwargs) -> None:
        super().__init__(r"(.*)", **kwargs)


WILDCARD = RegexElement(r"(.*)")


class Pattern:
    """
    A class representing a single pattern for a Sentence/Utterance
    pattern is built out of elements that match a single word in an utterance

    """

    def __init__(self, *elements: ta.Union[PatternElement, str, tuple, list, set]):
        elements_normalized = []
        for e in elements:
            if isinstance(e, PatternElement):
                elements_normalized.append(e)
            elif isinstance(e, str):
                elements_normalized.append(Words(e))
            elif isinstance(e, (tuple, list, set)):
                elements_normalized.append(Words(*e))

        self.pattern = re.compile(
            r"\s*".join(self.wrap_element_in_group(e) for e in elements_normalized),
            re.IGNORECASE,
        )

    @staticmethod
    def wrap_element_in_group(e):
        e_name_regex = ""
        if e.name:
            e_name_regex = f"?P<{e.name}>"
        return f"({e_name_regex}{e.regex_transformation()})"

    def __call__(self, user_input):
        return bool(self.pattern.fullmatch(user_input))

    def extract(self, user_input):
        m = self.pattern.fullmatch(user_input)
        if m:
            return m.groupdict()
        else:
            return {}


class Extractor(abc.ABC):
    def __init__(
        self, *patterns: Pattern, preprocess_func: ta.Callable[[str], str] = None
    ) -> None:
        self.patterns: ta.Tuple[Pattern, ...] = patterns
        if preprocess_func:
            self.preprocess_func = preprocess_func
        else:
            self.preprocess_func = lambda s: s


class Intent(Extractor):
    """
    Intent reposesents a group of patterns to match against an utterance
    """

    def __call__(self, user_input) -> bool:
        prepro_user_input = self.preprocess_func(user_input)
        for p in self.patterns:
            if p(prepro_user_input):
                return True
        return False


class Slots(Extractor):
    def __call__(self, user_input) -> dict:
        prepro_user_input = self.preprocess_func(user_input)
        slots = defaultdict(lambda: [])
        for p in self.patterns:
            for slot_name, slot_value in p.extract(prepro_user_input).items():
                slots[slot_name].append(slot_value)
        return dict(slots)


def extract_slot(extractor: Slots, user_input: str, slot_name: str) -> Optional[str]:
    slot_values = extractor(user_input).get(slot_name, [])
    if len(slot_values) > 0:
        return slot_values[0]
    return None


def intent_slot(
    *patterns: Pattern, preprocess_func: ta.Callable[[str], str] = None
) -> ta.Tuple[Intent, Slots]:
    """
        convenience wrapper to get intent+slot extractor from a set of patterns and preprocessor

    Keyword Arguments:
        preprocess_func {ta.Callable[[str], str]} -- preprocess function (default: {None})

    Returns:
        ta.Tuple[Intent, SlotsExtractor] -- intent and slot extractor
    """
    return Intent(*patterns, preprocess_func=preprocess_func), Slots(*patterns)



from typing import List, Optional
import enum
import json
import random

from pydantic import BaseModel


import coco.async_api as coco_sdk
from coco import ccml

import agt
from agt.state import OutOfContext, Outputs, ConversationState
from agt.nlu.word_regex import Intent, Pattern, WILDCARD

from coco.config_models import ActionsConfig, BlueprintConfig

available_intents = {
    "yes": Intent(
        Pattern(
            WILDCARD,
            (
                "yes",
                "yea",
                "yup",
                "sure",
                "ok",
                "of course",
                "yeah",
                "okay",
                "yep",
                "I do",
            ),
            WILDCARD,
        ),
        Pattern(("y", "totally", "naturally", "k")),
    ),
    "no": Intent(
        Pattern(
            WILDCARD,
            (
                "no",
                "nope",
                "never",
                "no way",
                "nah",
                "neh",
                "nay",
                "nop",
                "noo",
                "nooo",
            ),
            WILDCARD,
        ),
        Pattern("n"),
    ),
    "continue": Intent(
        Pattern(WILDCARD, ("yes", "yea", "yup", "sure", "ok"), WILDCARD),
        Pattern(("and", "go on")),
    ),
    "stop": Intent(
        Pattern(("stop", "bye", "this is boring")),
    ),
    "wildcard": Intent(Pattern(WILDCARD)),
}

AvailableIntents = enum.Enum(
    "AvailableIntents", zip(available_intents.keys(), available_intents.keys())
)


class OneTurnFollowup(BaseModel):
    intent_name: Optional[AvailableIntents]
    keywords: Optional[List[str]]
    followup_response: str


class OneTurnConfig(BlueprintConfig):
    oneturn_followups: List[OneTurnFollowup]
    available_intents: List[AvailableIntents] = list(AvailableIntents)


FOLLOWUP_CONFIG = json.loads(
    OneTurnConfig(
        blueprint_id="oneturn_followup",
        oneturn_followups=[
            OneTurnFollowup(intent_name=AvailableIntents.yes, followup_response="good"),
            OneTurnFollowup(
                intent_name=AvailableIntents.no, followup_response="but why?!"
            ),
            OneTurnFollowup(
                intent_name=AvailableIntents.stop, followup_response="ok bye!"
            ),
            OneTurnFollowup(keywords=["vanilla"], followup_response="good choice"),
        ],
    ).json()
)


def clean_keywords(keywords: List[str]) -> List[str]:
    return [k.strip() for k in keywords if isinstance(k, str) and len(k.strip()) > 0]


async def oneturn_followup(state: agt.ConversationState, config=FOLLOWUP_CONFIG):
    config_model: OneTurnConfig = OneTurnConfig.validate(config)
    user_input = await state.user_input()
    for fu in config_model.oneturn_followups:
        if fu.intent_name and available_intents[fu.intent_name.name](user_input):
            await state.say(fu.followup_response)
            return Outputs(control=fu.intent_name.name)
        if fu.keywords and Intent(
            Pattern(WILDCARD, clean_keywords(fu.keywords), WILDCARD)
        )(user_input):
            await state.say(fu.followup_response)
            return Outputs(control=fu.keywords[0])
    await state.out_of_context(user_input)


SAY_CONFIG = ActionsConfig(
    blueprint_id="oneturn_say", action_config={"line": ["This is a line"]}
).dict()


async def oneturn_say(state: agt.ConversationState, config=SAY_CONFIG):
    config_model: ActionsConfig = ActionsConfig.validate(config)
    await state.say(config_model.action_config["line"][0])


async def oneturn_say_v2(state: agt.ConversationState, line="This is a line", **kwargs):
    text = ccml.parse.clear_xml_tags(line)
    await state.say(text=text, ssml=line)


async def oneturn_say_v3(state: agt.ConversationState, lines=["This is a line"]):
    choosen_line = random.choice(lines)
    text = ccml.parse.clear_xml_tags(choosen_line)
    await state.say(text=text, ssml=choosen_line)


async def oneturn_followup_v2(
    state: agt.ConversationState, followups: List[OneTurnFollowup] = []
):
    user_input = await state.user_input()
    for fu_dict in followups:
        fu = OneTurnFollowup.validate(fu_dict)
        if fu.intent_name and available_intents[fu.intent_name.name](user_input):
            await state.say(fu.followup_response)
            return Outputs(control=fu.intent_name.name)
        if fu.keywords and Intent(
            Pattern(WILDCARD, clean_keywords(fu.keywords), WILDCARD)
        )(user_input):
            await state.say(fu.followup_response)
            return Outputs(control=fu.keywords[0])
    await state.out_of_context(user_input)


async def oneturn_say_followup(
    state: agt.ConversationState,
    line="This is a line",
    followups: List[OneTurnFollowup] = [],
):
    await state.say(line)
    return await oneturn_followup_v2(state, followups=followups)


async def save_to_context(state: agt.ConversationState, **params):
    state.memory.update(params)


async def aggregate_context(state: agt.ConversationState, **params):
    for k, v in params.items():
        if not isinstance(state.memory.get(k, []), list):
            state.memory[k] = [state.memory[k]]
        state.memory[k] = state.memory.get(k, []) + [v]


async def input_aggregator(state: agt.ConversationState, target_param="agg_inputs"):
    user_input = await state.user_input()  # not actually waiting
    state.memory[target_param] = state.memory.get(target_param, []) + [user_input]


async def oneliner_fallback(state, user_input):
    await state.say("Sorry I didn't get this")


async def example(state):
    await oneturn_say(state)
    with OutOfContext(state, oneliner_fallback):
        await oneturn_followup(state)


class Branch(BaseModel):
    branch_id: str
    intent_name: Optional[str]
    keywords: Optional[List[str]]


async def navigation(
    state: ConversationState, user_input=None, branches: List[Branch] = [], **kwargs
):
    user_input = user_input or await state.user_input()

    classic_intent_names = list(
        map(
            lambda b: b.get("intent_name"),
            filter(lambda b: "intent_name" in b and b["intent_name"]not in available_intents, branches),
        )
    )

    classic_intents_results = await coco_sdk.query_intents(
        intent_names=classic_intent_names, query=user_input
    )

    classic_intents_map = {r.name: r.result for r in classic_intents_results}

    for branch in branches:
        b = Branch.validate(branch)
        if (
            b.intent_name
            and b.intent_name in available_intents
            and available_intents[b.intent_name](user_input)
        ):
            return Outputs(control=b.branch_id)
        elif b.intent_name and classic_intents_map.get(b.intent_name):
            return Outputs(control=b.branch_id)
        if b.keywords and Intent(
            Pattern(WILDCARD, clean_keywords(b.keywords), WILDCARD)
        )(user_input):
            return Outputs(control=b.branch_id)

    await state.out_of_context(user_input)
